"""
beachfront-py
https://github.com/venicegeo/beachfront-py

Copyright 2016, RadiantBlue Technologies, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
"""
import os
import numpy
import requests
from gippy import GeoImage


def create_image(fout='', xsz=10, ysz=10, nodata=99, empty=False):
    """ Create test image, empty means all nodata, otherwise 0  """
    geoimg = GeoImage.create(fout, xsz=xsz, ysz=ysz, dtype='byte')
    geoimg.set_nodata(nodata)
    if not empty:
        arr = numpy.zeros((geoimg.xsize(), geoimg.ysize()), dtype='uint8')
        geoimg[0].write(arr)
    return geoimg


def download_image(url):
    """ Download a test image """
    fout = os.path.join(os.path.dirname(__file__), os.path.basename(url))
    if not os.path.exists(fout):
        print('Downloading image %s' % fout)
        stream = requests.get(url, stream=True)
        try:
            with open(fout, 'wb') as f:
                for chunk in stream.iter_content(1024):
                    f.write(chunk)
        except:
            raise Exception("Problem downloading %s" % url)
    return GeoImage(fout)



"""
beachfront-py
https://github.com/venicegeo/beachfront-py

Copyright 2016, RadiantBlue Technologies, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
"""
__version__ = '0.2.5'



"""
beachfront-py
https://github.com/venicegeo/beachfront-py

Copyright 2016, RadiantBlue Technologies, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
"""
from __future__ import print_function
import unittest
import os
from beachfront import process
from beachfront.logger import init_logger
from .utils import create_image, download_image


class TestProcess(unittest.TestCase):
    """ Test masking functions """

    @classmethod
    def setUpClass(cls):
        """ Initialize logging """
        init_logger(muted=True)

    def create_bimodal_image(self):
        """ Create image with a bimodal distribution """
        geoimg = create_image()
        arr = geoimg.read()
        arr[0:3, 0:10] = 1
        arr[3:5, 0:10] = 2
        arr[5:7, 0:10] = 9
        arr[7:10, 0:10] = 10
        geoimg[0].write(arr)
        return geoimg

    def test_threshold(self):
        """ Threshold binary image """
        geoimg = self.create_bimodal_image()
        threshold = round(process.otsu_threshold(geoimg[0]))
        self.assertEqual(threshold, 2)

        # add some nodata regions
        arr = geoimg.read()
        arr[4:6, 0:10] = geoimg[0].nodata()
        geoimg[0].write(arr)
        # test otsu threshold with nodata, should be similar since
        # nodata regions were an even number of 2's and 9's that cancel
        threshold = round(process.otsu_threshold(geoimg[0]))
        self.assertEqual(threshold, 2)

    def test_real_image(self):
        """ Threshold a real image """
        url = 'http://landsat-pds.s3.amazonaws.com/L8/139/045/LC81390452014295LGN00/LC81390452014295LGN00_B3.TIF'
        geoimg = download_image(url)
        geoimg.set_nodata(0)

        threshold = process.otsu_threshold(geoimg[0])
        geoimg[0] = geoimg[0] > threshold

        ext = os.path.splitext(geoimg.filename())[1]
        fout = os.path.splitext(geoimg.filename())[0] + '_otsu' + ext
        geoimg.save(fout, dtype='uint8')
        print('\nPerform visual inspection on %s' % fout)



"""
beachfront-py
https://github.com/venicegeo/beachfront-py

Copyright 2016, RadiantBlue Technologies, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
"""
import numpy as np
import logging


logger = logging.getLogger(__name__)


def otsu_threshold(georaster):
    """ Use Otsu's method to caluclate optimal threshold ignoring nodata values """
    nbins = 500
    logger.info('Reading from file %s' % georaster.filename(), action='Read file',
                actee=georaster.filename(), actor=__name__)
    hist = georaster.histogram(bins=nbins).astype(float)

    stats = georaster.stats()
    gain = (stats[1] - stats[0]) / nbins
    bin_centers = [stats[0] + i * gain for i in range(0, nbins)]

    # class probabilities for all possible thresholds
    weight1 = np.cumsum(hist)
    weight2 = np.cumsum(hist[::-1])[::-1]
    # class means for all possible thresholds
    mean1 = np.cumsum(hist * bin_centers) / weight1
    mean2 = (np.cumsum((hist * bin_centers)[::-1]) / weight2[::-1])[::-1]

    # Clip ends to align class 1 and class 2 variables:
    # The last value of `weight1`/`mean1` should pair with zero values in
    # `weight2`/`mean2`, which do not exist.
    variance12 = weight1[:-1] * weight2[1:] * (mean1[:-1] - mean2[1:]) ** 2

    idx = np.argmax(variance12)
    threshold = bin_centers[:-1][idx]

    return threshold



# Copyright 2016, RadiantBlue Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy of the
# License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import os
import sys
import datetime
import logging.config

FORMAT = ('<%(PRI)s>1 %(TIMESTAMP)s %(HOSTNAME)s %(APP_NAME)s %(process)s %(MSG_ID)s %(SD_ELEMENT)s'
          '(%(name)s:%(funcName)s) %(levelname)s %(message)s')

PRI_CODES = {
    'FATAL':    0,
    'CRITICAL': 2,
    'ERROR':    3,
    'WARN':     4,
    'WARNING':  4,
    'NOTICE':   5,
    'INFO':     6,
    'DEBUG':    7,
}


def init_logger(logLevel=logging.DEBUG, muted=False):
    logging.basicConfig(
        format=FORMAT,
        level=logLevel,
        stream=sys.stdout
    )

    logging.setLoggerClass(AuditableLogger)
    logger = logging.getLogger(__name__)
    if muted:
        logging.root.handlers = [logging.NullHandler()]

    # overall logging level
    logger.setLevel(1)
    return logger


def mute_logger():
    """ Mute the logger """
    logging.root.handlers = [logging.NullHandler()]


class AuditableLogger(logging.Logger):
    def _log(self, level, msg, args, exc_info=None, extra=None,
             actee='', action='', actor='', **kwargs):
        sd_params = []
        if actor:
            sd_params.append('actor="{}"'.format(actor))
        if action:
            sd_params.append('action="{}"'.format(action))
        if actee:
            sd_params.append('actee="{}"'.format(os.path.abspath(actee)))

        sd_element = ''
        if sd_params:
            sd_element = '[{SD_ID} {SD_PARAMS}] '.format(
                SD_ID='bfaudit@48851',
                SD_PARAMS=' '.join(sd_params),
            )

        pri = (1 << 3) | PRI_CODES.get(logging.getLevelName(level), PRI_CODES['NOTICE'])

        # Assemble RFC 5424 elements
        extra = {
            'ACTEE':     actee,
            'ACTION':    action,
            'ACTOR':     actor,
            'APP_NAME':  __name__.split('.')[0],
            'HOSTNAME':  os.uname()[1].lower(),
            'MSG_ID':    '-',
            'PRI':       pri,
            'SD_ELEMENT': sd_element,
            'TIMESTAMP': datetime.datetime.utcnow().isoformat() + 'Z',
        }

        super(AuditableLogger, self)._log(level, msg, args, exc_info, extra)



"""
beachfront-py
https://github.com/venicegeo/beachfront-py

Copyright 2016, RadiantBlue Technologies, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from __future__ import print_function
import unittest
import os
import glob
import json
import numpy
from osgeo import ogr
import logging
from beachfront import vectorize
from beachfront.logger import init_logger
from .utils import create_image, download_image


logger = logging.getLogger(__name__)


class TestVectorize(unittest.TestCase):
    """ Test masking functions """

    test_url = 'http://landsat-pds.s3.amazonaws.com/L8/139/045/LC81390452014295LGN00/LC81390452014295LGN00_B3.TIF'

    def setUp(self):
        numpy.set_printoptions(precision=2)
        init_logger(muted=True)

    def create_image_with_line(self):
        """ Create image with small line in middle """
        geoimg = create_image()
        arr = numpy.zeros((geoimg.xsize(), geoimg.ysize()), dtype='uint8')
        # arr[40:60, 50] = 1
        arr[2:8, 5] = 1
        geoimg[0].write(arr)
        self.truth_coords = [
            [50.36996467799592, 50.0], [50.39062402284878, 46.0953125],
            [50.530935406641134, 43.548242187499994], [50.66263873007813, 43.83125],
            [50.73724191982458, 44.75], [50.73724191982458, 44.75],
            [50.93924440283034, 50.0], [50.73724191982458, 55.25],
            [50.73724191982458, 55.25], [50.66263873007813, 56.16875],
            [50.530935406641134, 56.451757812500006], [50.39062402284878, 53.9046875], [50.36996467799592, 50.0]
        ]
        return geoimg

    def create_image_with_box(self):
        """ Create image with box in middle """
        geoimg = create_image(xsz=20, ysz=20)
        arr = numpy.zeros((geoimg.xsize(), geoimg.ysize()), dtype='uint8')
        # arr[25:75, 25:75] = 1
        arr[4:16, 4:16] = 1
        geoimg[0].write(arr)
        self.truth_coords = [
            [0.2, 0.8], [0.5, 0.8], [0.8, 0.8], [0.8, 0.5], [0.8, 0.19999999999999996],
            [0.5, 0.19999999999999996], [0.2, 0.19999999999999996], [0.2, 0.5]
            # [0.2, 0.8], [0.5, 0.8], [0.8, 0.8], [0.8, 0.5], [0.8, 0.2], [0.5, 0.2], [0.2, 0.2], [0.2, 0.5]
            # [0.25, 0.75], [0.5, 0.75], [0.75, 0.75], [0.75, 0.5], [0.75, 0.25], [0.5, 0.25], [0.25, 0.25], [0.25, 0.5]
        ]
        return geoimg

    def test_potrace_arrays(self):
        """ Trace numpy arrays using potrace """
        arr = numpy.zeros((10, 10), dtype='uint8')
        lines = vectorize.potrace_array(arr)
        self.assertEqual(type(lines), list)
        self.assertEqual(len(lines), 0)

    def test_save_geojson(self):
        """ Validate geojson returned from trace """
        geoimg = create_image()
        lines = vectorize.potrace(geoimg[0])
        fout = os.path.join(os.path.dirname(__file__), 'test.geojson')
        vectorize.save_geojson(lines, fout, source='test')
        with open(fout, 'r') as f:
            geojson = json.loads(f.read())
        self.assertTrue('type' in geojson.keys())
        self.assertEqual(geojson['type'], 'FeatureCollection')
        self.assertTrue('features' in geojson.keys())
        self.assertEqual(len(geojson['features']), 0)
        os.remove(fout)
        # TODO - check populated geojson

    # potrace does not trace lines, it's outer edge of polygons
    def _test_trace_line(self):
        """ Trace image of line """
        geoimg = self.create_image_with_line()
        geojson = vectorize.potrace(geoimg[0], latlon=True)

        # check returned geometry
        coords = geojson['features'][0]['geometry']['coordinates']
        self.assertEqual(len(coords), len(self.truth_coords))
        for c in coords:
            self.assertTrue(c in self.truth_coords)

    def test_potrace_box(self):
        """ Trace image of box """
        geoimg = self.create_image_with_box()
        lines = vectorize.potrace(geoimg[0], latlon=False)
        # check returned geometry
        self.assertEqual(len(lines[0]), len(self.truth_coords))
        for c in lines[0]:
            self.assertTrue(c in self.truth_coords)

    def test_potrace_minsize(self):
        """ Trace line with minsize smaller and larger than box """
        geoimg = self.create_image_with_box()

        # check that minsize did not filter out box
        lines = vectorize.potrace(geoimg[0], minsize=100, latlon=True)
        self.assertEqual(len(lines[0]), len(self.truth_coords))

        # check that minsize does filter out box
        lines = vectorize.potrace(geoimg[0], minsize=200)
        self.assertEqual(len(lines), 0)

    def test_potrace_nodata(self):
        """ Trace image with nodata present """
        geoimg = self.create_image_with_box()
        arr = geoimg.read()
        # make a nodata region
        arr[0:10, 0:10] = geoimg[0].nodata()
        geoimg[0].write(arr)
        lines = vectorize.potrace(geoimg[0], latlon=False)
        self.assertEqual(len(lines), 1)
        self.assertEqual(len(lines[0]), 7)
        self.assertEqual(lines[0],
                         [[0.65, 0.8], [0.8, 0.8], [0.8, 0.5], [0.8, 0.19999999999999996],
                         [0.5, 0.19999999999999996], [0.2, 0.19999999999999996], [0.2, 0.35]])

    def test_potrace_empty_image(self):
        """ Trace image that is empty """
        geoimg = create_image(empty=True)
        lines = vectorize.potrace(geoimg[0], latlon=True)
        self.assertEqual(len(lines), 0)
        geoj = vectorize.to_geojson(lines)
        self.assertEqual(len(geoj['features']), 0)

    def test_potrace_only_nodata(self):
        """ Trace image that has only nodata """
        geoimg = create_image(empty=True)
        arr = geoimg.read()
        # make a nodata region
        arr[0:5, 0:5] = geoimg[0].nodata()
        geoimg[0].write(arr)
        lines = vectorize.potrace(geoimg[0], latlon=True)
        self.assertEqual(len(lines), 0)
        geoj = vectorize.to_geojson(lines)
        self.assertEqual(len(geoj['features']), 0)

    def test_potrace_image(self):
        """ Trace landsat using an arbitrary cutoff """
        # get a complex test image (landsat)
        geoimg = download_image(self.test_url)
        geoimg.set_nodata(0)
        lines = vectorize.potrace(geoimg[0] > 9500, close=0)
        self.assertEqual(len(lines), 16551)
        fout = os.path.splitext(os.path.join(os.path.dirname(__file__), os.path.basename(self.test_url)))[0] + '.geojson'
        vectorize.save_geojson(lines, fout)
        print('\nPerform visual inspection on %s' % fout)

    def test_simplify(self):
        """ Simplify GeoJSON geometries """
        geoimg = download_image(self.test_url)
        geoimg.set_nodata(0)
        lines = vectorize.potrace(geoimg[0] > 9500, close=0)
        fout = os.path.join(os.path.dirname(__file__), 'test.geojson')
        vectorize.save_geojson(lines, fout)

        # check file
        df = ogr.Open(fout)
        layer = df.GetLayer()
        self.assertEqual(layer.GetFeatureCount(), len(lines))
        geom = json.loads(layer.GetNextFeature().ExportToJson())
        self.assertEqual(len(geom['geometry']['coordinates']), len(lines[0]))
        df = None

        # simplify and check file
        vectorize.simplify(fout, tolerance=0.001)
        df = ogr.Open(fout)
        layer = df.GetLayer()
        geom = json.loads(layer.GetNextFeature().ExportToJson())
        self.assertEqual(len(geom['geometry']['coordinates']), 22)

    def test_close_line_strings(self):
        """ Close line strings """
        geoimg = self.create_image_with_box()
        lines = vectorize.potrace(geoimg[0], latlon=False, close=6)
        self.assertEqual(len(lines[0]), 9)

    def test_close_line_strings_real(self):
        """ Close real line strings """
        geoimg = download_image(self.test_url)
        geoimg.set_nodata(0)
        lines = vectorize.potrace(geoimg[0] > 9500, close=5)
        fout = os.path.splitext(os.path.join(os.path.dirname(__file__), os.path.basename(self.test_url)))[0] + '_closed.geojson'
        vectorize.save_geojson(lines, fout)
        print('\nPerform visual inspection on %s' % fout)

    def test_antimeridian(self):
        """ Split line strings crossing antimeridian into two linestrings """
        fname = os.path.join(os.path.dirname(__file__), 'test-antimeridian.geojson')
        with open(fname) as f:
            gj = json.loads(f.read())
        lines = []
        for f in gj['features']:
            lines.append(f['geometry']['coordinates'])
        self.assertEqual(len(lines), 3)
        lines = vectorize.antimeridian_linesplit(lines)
        self.assertEqual(len(lines), 9)



"""
beachfront-py
https://github.com/venicegeo/beachfront-py

Copyright 2016, RadiantBlue Technologies, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
"""
import json
import numpy
from osgeo import osr, ogr
import potrace as _potrace
from pyproj import Proj, transform
import logging


logger = logging.getLogger(__name__)


def lines_to_features(lines, source='imagery'):
    """ Create features from lines """
    gid = 0
    features = []
    for line in lines:
        feature = {
            'type': 'Feature',
            'geometry': {
                'type': 'LineString',
                'coordinates': line
            },
            'properties': {
                'id': gid,
                'source': source
            }
        }
        features.append(feature)
        gid += 1
    return features


def to_geojson(lines, source='imagery'):
    geojson = {
        'type': 'FeatureCollection',
        'features': lines_to_features(lines, source=source),
    }
    return geojson


def save_geojson(lines, fout, source='imagery'):
    """ Save lines as GeoJSON file """
    geojson = to_geojson(lines, source=source)
    logger.info('Saving to file %s' % fout, action='Save file', actee=fout, actor=__name__)
    with open(fout, 'w') as f:
        f.write(json.dumps(geojson))
    return fout


def potrace_array(arr, minsize=10.0, tolerance=0.2, alphamax=0.0, opticurve=1):
    """ Trace numpy array using potrace """
    # tolerance, alphamax, and opticurve are best kept at their defaults
    bmp = _potrace.Bitmap(arr)
    logger.debug('potrace: minsize=%s, tolerance=%s, alphamax=%s, opticurve=%s' %
                 (minsize, tolerance, alphamax, opticurve))
    polines = bmp.trace(turdsize=minsize, turnpolicy=_potrace.TURNPOLICY_WHITE,
                        alphamax=alphamax, opticurve=opticurve, opttolerance=tolerance)
    lines = []
    for line in polines:
        lines.append(line.tesselate().tolist())

    return lines


def filter_nodata_lines(lines, mask, dist=3):
    """ Remove nodes within dist pixels of nodata regions or scene edges, splitting lines as needed  """
    # if mask.max() == 0:
    #    raise Exception('Empty mask!')
    newlines = []
    logger.debug('%s lines before filtering and splitting' % len(lines))
    xsize = mask.shape[0]
    ysize = mask.shape[1]
    minloc = 2
    xmaxloc = xsize - minloc - 1
    ymaxloc = ysize - minloc - 1
    for line in lines:
        startloc = 0
        for loc in range(0, len(line)):
            # check if this is masked point
            locx = int(line[loc][1])
            locy = int(line[loc][0])
            m = mask[max(locx-dist, 0):min(locx+dist, mask.shape[0]),
                     max(locy-dist, 0):min(locy+dist, mask.shape[1])]
            if m.sum() or (locx < minloc) or (locy < minloc) or (locx > xmaxloc) or (locy > ymaxloc):
                if (loc-startloc) > 1:
                    newlines.append(line[startloc:loc])
                startloc = loc + 1
        # add the last segment
        if (loc-startloc) > 1:
            newlines.append(line[startloc:])
    logger.debug('%s lines after filtering and splitting' % len(newlines))
    return newlines


def close_line_strings(lines, dist=5.0):
    """ Close line strings (1st pt = last pt) if within dist (in pixels) """
    if dist == 0:
        return lines
    dists = []
    for line in lines:
        xdist = line[-1][0] - line[0][0]
        ydist = line[-1][1] - line[0][1]
        d = numpy.sqrt(xdist**2 + ydist**2)
        dists.append(d)
        if d <= dist:
            line.append(line[0])
    if len(dists) > 0:
        logger.debug('Endpoint distances: min=%s, max=%s, mean=%s' % (min(dists), max(dists), numpy.mean(dists)))
    return lines


def antimeridian_linesplit(lines):
    """ Split lines crossing the 180th meridian """
    newlines = []
    for line in lines:
        lastpt = [line[0][0], line[0][1]]
        newline = [lastpt]
        for point in line[1:]:
            pt = [point[0], point[1]]
            # check if crosses the antimeridian
            if (abs(pt[0]) > 175) and ((lastpt[0] * pt[0]) < 0):
                xmin = max((lastpt[0], pt[0]))
                xmax = 360 + min((lastpt[0], pt[0]))
                # calculate crossing latitude
                ydist = max([pt[1], lastpt[1]]) - min([pt[1], lastpt[1]])
                m = ydist / (xmax-xmin)
                if pt[0] > lastpt[0]:
                    b = pt[1] - m * pt[0]
                else:
                    b = lastpt[1] - m * lastpt[0]
                latcross = m * 180.0 + b
                # cap off this line
                newline.append([180.0 * numpy.sign(lastpt[0]), latcross])
                newlines.append(newline)
                # create new next line
                newline = [[180.0 * numpy.sign(pt[0]), latcross]]
            newline.append(pt)
            lastpt = pt
        newlines.append(newline)
    return newlines


def convert_to_latlon(geoimg, lines):
    srs = osr.SpatialReference(geoimg.srs()).ExportToProj4()
    projin = Proj(srs)
    projout = Proj(init='epsg:4326')
    newlines = []
    for line in lines:
        l = []
        for point in line:
            pt = transform(projin, projout, point[0], point[1])
            l.append(pt)
        newlines.append(l)
    return antimeridian_linesplit(newlines)


def potrace(geoimg, latlon=True, close=5.0, **kwargs):
    """ Trace raster image using potrace and return geolocated or lat-lon coordinates """
    # assuming single band
    arr = geoimg.read()
    mask = geoimg.nodata_mask()

    arr[mask == 1] = 0
    lines = potrace_array(arr, **kwargs)
    lines = filter_nodata_lines(lines, mask)
    lines = close_line_strings(lines, dist=close)

    newlines = []
    for line in lines:
        newline = []
        for point in line:
            pt = geoimg.geoloc(point[0], point[1])
            newline.append([pt.x(), pt.y()])
        newlines.append(newline)
    if latlon:
        newlines = convert_to_latlon(geoimg, newlines)
    return newlines


def simplify(filename, tolerance=0.00035):
    """ Simplify GeoJSON vector """
    # tolerance is set using GeoJSON map units. Expects decimal degrees, but should work with anything
    if tolerance is None:
        return filename
    logger.info('Updating file %s with simplified geometries' % filename, action='Updating file',
                actee=filename, actor=__name__)
    vs = ogr.Open(filename, 1)  # 1 opens the file in read/write mode, 0 for read-only mode
    layer = vs.GetLayer()
    feat = layer.GetNextFeature()
    while feat is not None:
        geo = feat.geometry()
        simple = geo.Simplify(tolerance)
        feat.SetGeometry(simple)
        layer.SetFeature(feat)
        feat = layer.GetNextFeature()
    layer = None
    vs.Destroy()



"""
beachfront-py
https://github.com/venicegeo/beachfront-py

Copyright 2016, RadiantBlue Technologies, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
"""
import unittest
import os
from gippy import Options
from beachfront import mask
from .utils import download_image
from beachfront.logger import init_logger


class TestMask(unittest.TestCase):
    """ Test masking functions """

    # some cirrus clouds
    imgurl = 'http://landsat-pds.s3.amazonaws.com/L8/008/028/LC80080282016215LGN00/LC80080282016215LGN00_B3.TIF'
    qimgurl = 'http://landsat-pds.s3.amazonaws.com/L8/008/028/LC80080282016215LGN00/LC80080282016215LGN00_BQA.TIF'

    vfilename = os.path.join(os.path.dirname(__file__), 'test-coast.geojson')

    #bbox = [-78.5964604299, -11.1797039494, -76.5056457676, -9.06874079313]
    bbox = [-65.1126692276, 44.929454947, -62.0891890425, 47.100321387]

    @classmethod
    def setUpClass(cls):
        """ Initialize logging """
        init_logger(muted=True)

    def setUp(self):
        Options.set_verbose(2)

    def test_open_vector(self):
        """ Open WFS and check number of features """
        ds, layer = mask.open_vector(self.vfilename)
        self.assertEqual(layer.GetFeatureCount(), 16)

    def test_get_features(self):
        """ Open WFS using bounding box """
        ds, layer = mask.open_vector(self.vfilename)
        features = mask.get_features_as_geojson(layer, bbox=self.bbox)
        self.assertEqual(layer.GetFeatureCount(), 14)

    def test_get_features_unioned(self):
        """ Get union of all features within bounding box """
        ds, layer = mask.open_vector(self.vfilename)
        poly = mask.get_features_as_geojson(layer, bbox=self.bbox, union=True)
        self.assertAlmostEqual(poly['coordinates'][0][0][0], -65.0, 4)

    def test_mask_with_vector(self):
        """ Mask image with a WFS using gippy cookie_cutter """
        geoimg = download_image(self.imgurl)
        geoimg.set_nodata(0)
        imgout = mask.mask_with_vector(geoimg, (self.vfilename, ''))
        self.assertEqual(imgout.nbands(), 1)

    def _test_mask(self):
        """ Create mask image from bitwise mask """
        bqaimg = download_image(self.qimgurl)
        maskimg = mask.create_mask_from_bitmask(bqaimg)
        arr = maskimg.read()
        self.assertTrue(arr.sum(), 23096984)



"""
beachfront-py
https://github.com/venicegeo/beachfront-py

Copyright 2016, RadiantBlue Technologies, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
"""
import os
import tempfile
from osgeo import ogr
from gippy import GeoImage, GeoVector
import gippy.algorithms as alg
import json
import numpy as np
import logging


logger = logging.getLogger(__name__)


def open_vector(filename, layer=''):
    """ Fetch wfs features within bounding box """
    logger.info('Opening %s as vector file' % filename, action='Open file', actee=filename, actor=__name__)
    ds = ogr.Open(filename)
    if layer == '':
        layer = ds.GetLayer(0)
    else:
        layer = ds.GetLayer(layer)
    return (ds, layer)


def get_features_as_geojson(layer, bbox=None, union=False):
    """ Get features in this layer and return as GeoJSON """
    if bbox is not None:
        layer.SetSpatialFilterRect(bbox[0], bbox[3], bbox[2], bbox[1])
    poly = ogr.Geometry(ogr.wkbPolygon)
    if union:
        for feature in layer:
            geom = feature.GetGeometryRef()
            #  required for ogr2
            if hasattr(geom, 'GetLinearGeometry'):
                geom = geom.GetLinearGeometry()
            poly = poly.Union(geom)
        if bbox is not None:
            wkt = "POLYGON ((%s %s, %s %s, %s %s, %s %s, %s %s))" % \
                (bbox[0], bbox[1], bbox[2], bbox[1], bbox[2], bbox[3], bbox[0], bbox[3], bbox[0], bbox[1])
            bbox_wkt = ogr.CreateGeometryFromWkt(wkt)
            poly = poly.Intersection(bbox_wkt)
    return json.loads(poly.ExportToJson())


def get_features(layer, bbox=None, union=False, filename=''):
    """ Get features in this layer and return as GeoVector """
    features = get_features_as_geojson(layer, bbox=bbox, union=union)
    if filename == '':
        f, filename = tempfile.mkstemp(suffix='.geojson')
        logger.info('Saving JSON as vector file', action='Save file', actee=filename, actor=__name__)
        # if no geometries, returns empty geometry collection
        if features['type'] == 'GeometryCollection':
            features = {'type': 'FeatureCollection', 'features': []}
        os.write(f, json.dumps(features))
        os.close(f)
    else:
        logger.info('Saving JSON as vector file', action='Save file', actee=filename, actor=__name__)
        logger.info('Writing GeoJSON to file %s' % filename)
        with open(filename, 'w') as f:
            f.write(json.dumps(features))
    # create GeoVector
    return GeoVector(filename)


def get_coastline(bbox):
    """ Get coastline GeoJSON within bounding box """
    cmask = open_vector(os.path.join(os.path.dirname(__file__), 'coastline.shp'))
    lons = [c[0] for c in bbox['features'][0]['geometry']['coordinates'][0]]
    lats = [c[1] for c in bbox['features'][0]['geometry']['coordinates'][0]]
    bbox = [min(lons), min(lats), max(lons), max(lats)]
    gj = get_features_as_geojson(cmask[1], bbox=bbox, union=True)
    return gj


def mask_with_vector(geoimg, vector, filename=''):
    """ Mask geoimage with a vector """
    ext = geoimg.geo_extent()
    ds, layer = open_vector(vector[0], vector[1])

    geovec = get_features(layer, bbox=[ext.x0(), ext.y0(), ext.x1(), ext.y1()], union=True)
    if geovec.nfeatures() == 0:
        raise RuntimeError('No features after masking')

    res = geoimg.resolution()
    logger.info('Saving to file %s' % filename, action='Save file', actee=filename, actor=__name__)
    imgout = alg.cookie_cutter([geoimg], filename=filename, feature=geovec[0],
                               proj=geoimg.srs(), xres=res.x(), yres=res.y())
    return imgout


def create_mask_from_bitmask(geoimg, filename=''):
    """ Mask geoimg with a series of provided bitmasks """
    # medium and high confidence clouds
    nodata = int('0000000000000001', 2)
    clouds = int('1000000000000000', 2)
    cirrus = int('0011000000000000', 2)

    # calculate mask
    arr = geoimg.read().astype('int16')
    # it is a good data mask
    mask = (np.bitwise_and(arr, nodata) != nodata) & \
           (np.bitwise_and(arr, clouds) < clouds) & \
           (np.bitwise_and(arr, cirrus) < cirrus)

    # create mask file
    logger.info('Saving to file %s' % filename, action='Save file', actee=filename, actor=__name__)
    maskimg = GeoImage.create_from(geoimg, filename=filename, dtype='uint8')
    maskimg.set_nodata(0)
    maskimg[0].write(mask.astype('uint8'))

    return maskimg



"""
beachfront-py
https://github.com/venicegeo/beachfront-py

Copyright 2016, RadiantBlue Technologies, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
"""
import logging
import unittest
from beachfront.logger import init_logger


class TestLoggers(unittest.TestCase):
    ''' Test Cumulus logger '''

    def setUp(self):
        """ Clear root handlers """
        logging.root.handlers = []

    def test_config_null(self):
        """ Check configuration of muted logger """
        init_logger(muted=True)
        # null handler
        logger = logging.getLogger(__name__)
        self.assertEqual(len(logger.root.handlers), 1)
        self.assertTrue(isinstance(logger.root.handlers[0], logging.NullHandler))

    def test_config_stdout(self):
        """ Check configuration of logger with stdout stream """
        init_logger()
        logger = logging.getLogger(__name__)
        self.assertEqual(len(logger.root.handlers), 1)

        self.assertTrue(isinstance(logger.root.handlers[0], logging.StreamHandler))

    def _test_logging(self):
        """ Log messages """
        init_logger()
        logger = logging.getLogger(__name__)
        logger.info('logging test')



#!/usr/bin/env python
"""
beachfront-py
https://github.com/venicegeo/beachfront-py

Copyright 2016, RadiantBlue Technologies, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
"""

import os
from codecs import open
from setuptools import setup, find_packages
import imp

__version__ = imp.load_source('beachfront.version', 'beachfront/version.py').__version__

here = os.path.abspath(os.path.dirname(__file__))

# get the dependencies and installs
with open(os.path.join(here, 'requirements.txt'), encoding='utf-8') as f:
    all_reqs = f.read().split('\n')

install_requires = [x.strip() for x in all_reqs if 'git+' not in x]

setup(
    name='beachfront',
    version=__version__,
    description='library extracting coastline regions from raster data',
    author='Matthew Hanson (matthewhanson)',
    license='GPL',
    url='https://github.com/venicegeo/beachfront-py',
    classifiers=[
        'Framework :: Pytest',
        'Topic :: Scientific/Engineering :: GIS',
        'Topic :: Scientific/Engineering',
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 2.7',
    ],
    packages=find_packages(exclude=['docs', 'tests*']),
    include_package_data=True,
    install_requires=install_requires
)
